"""Test POST/PUT/PATCH/DELETE du client d'API."""
import json
import pytest
from scrippy_api.api import Client
from scrippy_api import ScrippyApiError


def test_post():
  """POST."""
  data = {"user": {"first_name": "Harry", "last_name": "Fink", "password": "D3ADP4RR0T"}}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  response = client.request(method="POST", url=url, data=json.dumps(data))
  assert response.status_code == 200


def test_bad_post():
  """POST."""
  data = {"user": {"parrot": "dead"}}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  with pytest.raises(ScrippyApiError) as err:
    response = client.request(method="POST", url=url, data=json.dumps(data))
  assert "[HTTPError] 400 Client Error: Bad Request for url: http://gunicorn:8080/user" in str(err.value)


def test_put():
  """PUT."""
  data = {"user": {"first_name": "Harry", "last_name": "Fink", "password": "D3ADP4RR0T"}}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  response = client.request(method="PUT", url=url, data=json.dumps(data))
  assert response.status_code == 200


def test_bad_put():
  """PUT."""
  data = {"user": {"parrot": "dead"}}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  with pytest.raises(ScrippyApiError) as err:
    response = client.request(method="PUT", url=url, data=json.dumps(data))
  assert "[HTTPError] 400 Client Error: Bad Request for url: http://gunicorn:8080/user" in str(err.value)


def test_patch():
  """PATCH."""
  data = {"user": {"first_name": "Harry", "last_name": "Fink", "password": "D3ADP4RR0T"}}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  response = client.request(method="PATCH", url=url, data=json.dumps(data))
  assert response.status_code == 200


def test_bad_patch():
  """PUT."""
  data = {"user": {"parrot": "dead"}}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  with pytest.raises(ScrippyApiError) as err:
    response = client.request(method="PATCH", url=url, data=json.dumps(data))
  assert "[HTTPError] 400 Client Error: Bad Request for url: http://gunicorn:8080/user" in str(err.value)


def test_delete():
  """DELETE."""
  data = {"user": {"first_name": "Harry", "last_name": "Fink", "password": "D3ADP4RR0T"}}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  response = client.request(method="DELETE", url=url, data=json.dumps(data))
  assert response.status_code == 200


def test_bad_delete():
  """PUT."""
  data = {"user": {"parrot": "dead"}}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  with pytest.raises(ScrippyApiError) as err:
    response = client.request(method="DELETE", url=url, data=json.dumps(data))
  assert "[HTTPError] 400 Client Error: Bad Request for url: http://gunicorn:8080/user" in str(err.value)


def test_delete_unknown_user():
  """PUT."""
  data = {"user": {"first_name": "Luiggi", "last_name": "Vercotti", "password": "SP4N1SH1NQU1S1T10N"}}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  with pytest.raises(ScrippyApiError) as err:
    response = client.request(method="DELETE", url=url, data=json.dumps(data))
  assert "Request error: [HTTPError] 404 Client Error: Not Found for url: http://gunicorn:8080/user" in str(err.value)



import logging
import jsonschema


def validate(instance, schema):
  logging.debug("[+] Validating parameters")
  try:
    jsonschema.validate(instance=instance, schema=schema)
  except Exception:
    logging.debug(" '-> Invalid parameters")
    return False
  logging.debug(" '-> Valid parameters")
  return True



"""Test de base du client d'API."""
import os
import pytest
from scrippy_api.api import Client
from scrippy_api import ScrippyApiError


def test_get():
  """Connection basique."""
  params = {}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  response = client.request(method="GET", url=url, params=params)
  assert response.status_code == 200


def test_bad_get():
  """Connection basique."""
  params = {"parrot": "dead"}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  with pytest.raises(ScrippyApiError) as err:
    response = client.request(method="GET", url=url, params=params)
  assert "[HTTPError] 400 Client Error: Bad Request for url: http://gunicorn:8080/user?parrot=dead" in str(err.value)


def test_get_unknown_user():
  """Connection basique."""
  params = {"first_name": "Luiggi", "last_name": "Vercotti", "password": "SP4N1SH1NQU1S1T10N"}
  url = "http://gunicorn:8080/user"
  client = Client(verify=False)
  with pytest.raises(ScrippyApiError) as err:
    response = client.request(method="GET", url=url, params=params)
  assert "Request error: [HTTPError] 404 Client Error: Not Found for url: http://gunicorn:8080/user?first_name=Luiggi&last_name=Vercotti&password=SP4N1SH1NQU1S1T10N" in str(err.value)


def test_download():
  """Test de téléchargement. """
  url = "https://www.mcos.nc/user/pages/01.home/_sliders/01.illustration/banner.jpg"
  local_dir = "./"
  local_path = os.path.join(local_dir, url.split('/')[-1])
  client = Client()
  assert client.download(url, local_dir)
  assert os.path.isfile(local_path)



"""Module facilitant l'utilisation d'API ReST."""

import os
import logging
import requests
from scrippy_api import ScrippyApiError


class Client:
  """
  Classe permettant l'utilisation d'API REST.

  https://fr.wikipedia.org/wiki/Representational_state_transfer.
  """

  def __init__(self, verify=True, exit_on_error=True):
    """
    Initialise la classe Client.

    :param method
    :param: verify: Vérifie le certificat SSL, defaults to True
    :exit_on_error: Lève immédiatement une erreur ScrippyApiError en cas d'erreur lors de la requète, defaults to True
    """
    self.exit_on_error = exit_on_error
    self.verify = verify
    self.session = requests.Session()

  def download(self, url, local_path, filename=''):
    """
    Permet le téléchargement d'un fichier.

    :param url: L'UrL à partir de laquelle télécharger le fichier
    :param local_path: Le répertoire dans lequel sauvegarder le fichier
    :param filename: Le nom du fichier local
    """
    if len(filename) == 0:
      filename = url.split('/')[-1]
    filename = os.path.join(local_path, filename)
    logging.debug("[+] Downloading file")
    logging.debug(f" '-> From: {url}")
    logging.debug(f" '-> To: {filename}")
    try:
      with requests.get(url, stream=True) as response:
        response.raise_for_status()
        with open(filename, 'wb') as dl_file:
          for chunk in response.iter_content(chunk_size=8192):
            dl_file.write(chunk)
      return True
    except Exception as err:
      if self.exit_on_error:
        err_msg = f"Request error: [{err.__class__.__name__}] {err}"
        raise ScrippyApiError(err_msg) from err
      else:
        logging.warning(f"Request error: [{err.__class__.__name__}] {err}")
        return response

  def request(self, method, url, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=None, proxies=None, json=None):
    """
    Permet d'exécuter une requète HTTP.

    :param  method
    :param: url: L'URL à atteindre
    :param: params: Les paramètres (GET) de la requète, defaults to None
    :param: data: Les paramètres (POST) de la requète, defaults to None
    :param: headers: Les entêtes de la requète, defaults to None
    :param: cookies: Les cookies de la requète, defaults to None
    :param: file: Le fichier envoyé (POST), defaults to None
    :param: auth: Les informations d'authentification (BASIC AUTH) sous forme d'un tuple (user, password), defaults to None
    :param: timeout: Le délai maximum à attendre, defaults to None
    :param: proxies: La liste des serveurs mandataires utilisés, defaults to None
    :param: json: JSON à envoyer dans le corps de la requète (POST), defaults to None
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    logging.debug("[+] Sending request to server")
    default_get_kwargs = {"params": params,
                          "timeout": timeout,
                          "headers": headers,
                          "cookies": cookies,
                          "proxies": proxies,
                          "auth": auth,
                          "verify": self.verify}
    default_post_kwargs = {"data": data, "json": json, "files": files}
    # TO BE IMPLEMENTED:
    #  "CONNECT": self._connect,
    #  "OPTIONS": self._options,
    #  "TRACE": self._trace,
    #  "HEAD": self._trace,
    methods = {"GET": {"method": self._get,
                       "kwargs": default_get_kwargs},
               "POST": {"method": self._post,
                        "kwargs": {**default_get_kwargs, **default_post_kwargs}},
               "PUT": {"method": self._put,
                       "kwargs": {**default_get_kwargs, **default_post_kwargs}},
               "DELETE": {"method": self._delete,
                          "kwargs": {**default_get_kwargs, **default_post_kwargs}},
               "PATCH": {"method": self._patch,
                         "kwargs": {**default_get_kwargs, **default_post_kwargs}}}
    try:
      response = None
      response = methods[method]["method"](url, **methods[method]["kwargs"])
      response.raise_for_status()
      return response
    except Exception as err:
      if self.exit_on_error:
        err_msg = f"Request error: [{err.__class__.__name__}] {err}"
        raise ScrippyApiError(err_msg) from err
      else:
        logging.warning(f"Request error: [{err.__class__.__name__}] {err}")
        return response

  def _get(self, url, **kwargs):
    return self.session.get(url, **kwargs)

  def _post(self, url, **kwargs):
    return self.session.post(url, **kwargs)

  def _head(self, url, **kwargs):
    return self.session.head(url, **kwargs)

  def _put(self, url, **kwargs):
    return self.session.put(url, **kwargs)

  def _delete(self, url, **kwargs):
    return self.session.delete(url, **kwargs)

  def _patch(self, url, **kwargs):
    return self.session.patch(url, **kwargs)



"""Module de chargement de définition d'API."""

import yaml
import logging
from scrippy_api import ScrippyApiError


class ApiLoader:
  """L'objet ApiLoader permet le chargement d'une API définie à l'aide d'un fichier YAML."""

  def __init__(self):
    """Initialise la classe ApiLoader."""
    self.api = {}

  def _walk_api(self, dic, path):
    """Parcours la représentation de l'API passée en argument sous forme de dictionnaire."""
    for key, value in dic.items():
      if isinstance(value, list):
        for action_list in value:
          for action in action_list:
            ppath = ".".join(path)
            ppath = f"{ppath}.{action}"
            path.append(ppath)
            self.api[ppath] = action_list[action]
            path.pop()
      elif isinstance(value, dict):
        path.append(key)
        self._walk_api(value, path)
        path.pop()

  def load_api(self, api_definition):
    """Charge l'API à partir du fichier YAML passé en argument."""
    logging.debug("[+] Loading API")
    logging.debug(f" '-> {api_definition}")
    try:
      with open(api_definition, mode="r") as yaml_file:
        api_yaml = yaml.load(yaml_file, Loader=yaml.FullLoader)
      self._walk_api(dic=api_yaml, path=[])
    except Exception as err:
      err_msg = f"Unknown error: [{err.__class__.__name__}] {err}"
      raise ScrippyApiError(err_msg) from err

  def get_endpoint_info(self, endpoint):
    """
    Renvoi les informations du endpoint passé en argument.

    Les informations renvoyées sont un dictionnaire tel que:
    {"method": <HTTP METHOD>, "url": <URL>}
    """
    logging.debug(f"[+] Getting data for: {endpoint}")
    try:
      return self.api[endpoint]
    except KeyError as err:
      err_msg = f"Unknown endpoint: {endpoint}"
      raise ScrippyApiError(err_msg) from err
